﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace DeJong
{
    public class Chromosome : IComparable<Chromosome>
    {
        /// <summary>
        /// Konstruktor chromosomu liczba genów 2 - ponieważ występują dwie zmienne w
        /// II funkcji De Jonga - stąd l == 2
        /// </summary>
        /// <param name="n">Długość chromosomu</param>
        public Chromosome(int n)
        {
            table = new double[n];
            delta = new double[n];

            for (var i = 0; i < table.Length; i++)
            {
                table[i] = 0;
                delta[i] = 1;
            }

            // obliczanie współczynników mutacji
            ro = C / (Math.Pow(2.0 * Math.Pow((double)n, 0.5), 0.5));
            roPrim = C / (Math.Pow(2.0 * (double)n, 0.5));

        }

        /// <summary>
        /// Konstruktor kopiujący
        /// </summary>
        /// <param name="c">Chromosom do skopiowania</param>
        public Chromosome(Chromosome c)
        {
            this.fitness = c.fitness;

            if (c.table == null) this.table = null;
            else
            {
                this.table = new double[c.table.Length];
                for (var i = 0; i < this.table.Length; i++)
                {
                    this.table[i] = c.table[i];
                }
            }

            if (c.delta == null)
            {
                this.delta = null;
            }
            else
            {
                this.delta = new double[c.delta.Length];
                for (var i = 0; i < this.delta.Length; i++)
                {
                    this.delta[i] = c.delta[i];
                }
            }
        }

        /// <summary>
        /// Operacja mutacji chromosomu
        /// </summary>
        public void Mutation()
        {
            var random = new Random();

            var valA = RandomProvider.NextNormal();

            for (var i = 0; i < table.Length; i++)
            {
                var valB = RandomProvider.NextNormal();
                delta[i] = delta[i] * Math.Exp(roPrim * valA + ro * valB);
                table[i] = table[i] + delta[i] * valB;
            }
        }

        /// <summary>
        /// Operacja krzyżowania dwóch chromosomów
        /// </summary>
        /// <param name="chromosome">Chromosom do skrzyżowania z tym chromosomem</param>
        /// <returns>Wynik operacji krzyżowania</returns>
        public Chromosome Crossover(Chromosome chromosome)
        {
            var result = new Chromosome(this.table.Length);

            result.table = new double[this.table.Length];
            result.delta = new double[this.delta.Length];

            for (var i = 0; i < result.table.Length; i++)
            {
                result.table[i] = (this.table[i] + chromosome.table[i]) * 0.5d;
                result.delta[i] = (this.delta[i] + chromosome.delta[i]) * 0.5d;
            }

            return result;
        }

        /// <summary>
        /// Metoda obliczająca wartość przystosowania chromosomu
        /// </summary>
        public void Evalutate()
        {
            this.fitness = DeJong2nd(table);
        }

        /// <summary>
        /// II funkcja De Jonga
        /// </summary>
        /// <param name="value">tabela wartości</param>
        /// <returns>wartość II funkcji De Jonga</returns>
        private double DeJong2nd(double [] value) 
        {
            var sum = 0.0;

            for (var i = 0; i < value.Length - 1; i++)
            {
                sum += 100 * Math.Pow((Math.Pow(value[i], 2) - value[i + 1]), 2) + Math.Pow((1 - value[i]), 2);
            }

            return sum;
        }

        /// <summary>
        /// Wartość funkcji Fitness dla chromosomu
        /// </summary>
        public double Fitness { get { return fitness; } }

        /// <summary>
        /// Wartości rozwiązania reprezentowane przez chromosom
        /// </summary>
        public double[] Table { get { return table; } }

        #region IComparable<Chromosome> Members

        /// <summary>
        /// Porównuje dwa chromosomy
        /// </summary>
        /// <param name="other">Drugi chromosom do porównania</param>
        /// <returns>1 jeżeli ten chromosom jest lepszy, 0 jeżeli są równie dobre lub -1 jeżeli drugi chromosom jest lepszy</returns>
        public int CompareTo(Chromosome other)
        {
            if (other == null || fitness > other.fitness) return 1;
            if (fitness == other.fitness) return 0;
            return -1;
        }

        #endregion

        // stała w obliczeniach dla ro i roPrim
	    private static readonly double C = 1.0;

	    // wartość przysotowania chromosomu
	    private double fitness;

	    // tablica wartości bazowych w chromosomie
	    private double[] table;
	    // tablica odchyleń dla mutacji
	    private double[] delta;

	    // wartości pomocnicze wykorzystywane podczas obliczeń mutacji
	    private double ro;
	    private double roPrim;

        
    }
}
