﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;

namespace DeJong
{
    /// <summary>
    /// Implementacja strategi ewolucyjnej (mi + lambda) dla II funkcji De Jonga.
    /// </summary>
    public class EvolutionStrategy
    {
        public EvolutionStrategy()
        {
            
        }

        public void Run()
        {
            ReadProperties();

            RunAlgorithm(dimension, populationSize, reproductionSize);
        }

        private void PrintChromosome(Chromosome c)
        {
            for (var i = 0; i < c.Table.Length; i++)
            {
                Console.Write(i == 0 ? "[" : ",");
                Console.Write(c.Table[i]);
            }

            Console.Write("] - fitness: ");
            Console.WriteLine(c.Fitness);
        }

        private List<Chromosome> Initialization(int populationSize, int dimension)
        {
            var result = new List<Chromosome>();
            for (var i = 0; i < populationSize; i++)
            {
                result.Add(new Chromosome(dimension));
            }
            return result;
        }

        private void Evaluation(List<Chromosome> population)
        {
            foreach (var c in population) c.Evalutate();

            // posortuj osobniki wg wartośći przysotowania
            // szukamy min funkcji więc najlepszy osobnik to taki o najmniejszej
            // wartośći fitness - wartośc 2 funkcji De Jonga
            population.Sort();
        }

        private List<Chromosome> Reproduction(List<Chromosome> population, int reproductionSize)
        {
            var random = new Random();
            var newPopulation = new List<Chromosome>();

            for (var i = 0; i < reproductionSize; i++)
            {
                var parentA = population[random.Next(populationSize)];
                var parentB = population[random.Next(populationSize)];
                var descendant = parentA.Crossover(parentB);

                newPopulation.Add(descendant);
            }

            return newPopulation;
        }

        private List<Chromosome> Selection(List<Chromosome> population, List<Chromosome> newPopulation)
        {
            var populationSize = population.Count;

            var all = new List<Chromosome>();
            all.AddRange(population);
            all.AddRange(newPopulation);
            all.Sort();

            var result = new List<Chromosome>();
            for (var i = 0; i < populationSize; i++)
            {
                result.Add(new Chromosome(all[i]));
            }

            return result;
        }


        private void RunAlgorithm(int dimension, int populationSize, int reproductionSize)
        {
            var startTime = DateTime.Now;

            var iteration = 0;
            var stop = false;

            // inicjalizacja populacji początkowej
            var population = Initialization(populationSize, dimension);

            Evaluation(population);

            var theBest = population.First();
            var iterationWithoutChangeTheBest = 0;

            while (!stop)
            {
                var newPopulation = Reproduction(population, reproductionSize);
                
                foreach (var c in newPopulation) c.Mutation();

                Evaluation(newPopulation);

                population = Selection(population, newPopulation);

                population.Sort();

                var localTheBest = population.First();
                if (localTheBest.Fitness < theBest.Fitness)
                {
                    theBest = new Chromosome(localTheBest);
                    iterationWithoutChangeTheBest = 0;
                }
                else iterationWithoutChangeTheBest++;

                if (theBest.Fitness < 1.0E-10) stop = true;

                iteration++;
            }

            var time = (DateTime.Now - startTime).Milliseconds;
            var output = new string[] 
            {
                "Dane wejściowe -------------------------------------------------------------",
                string.Format("Wymiar funkcji testowej                 - {0}", dimension),
                string.Format("Liczba osobników populacji bazowej      - {0}", populationSize),
                string.Format("Liczba osobników populacji reprodukcji  - {0}", reproductionSize),
                "Wyniki ---------------------------------------------------------------------",
                string.Format("Liczba iteracji                         - {0}", iteration),
                string.Format("Czas trwania                            - {0} [ms]", time ),
                string.Format("Średni czas trwania iteracji            - {0} [ms]", time/iteration),
                "Rozwiązanie ----------------------------------------------------------------"
            };

            foreach (var s in output) Console.WriteLine(s);
            PrintChromosome(theBest);
        }

        private void ReadProperties()
        {
            try
            {
                var br = new StreamReader("properties.txt");

                dimension = int.Parse(br.ReadLine());
                populationSize = int.Parse(br.ReadLine());
                reproductionSize = int.Parse(br.ReadLine());

                br.Close();
            }
            catch (Exception e)
            {
                Console.WriteLine("Wystąpił błąd wczytywana ustawień algorytmu z pliku properties.txt");
                Console.WriteLine(e.StackTrace);
                Environment.Exit(1);
            }
        }

        // liczba wymiarów funkcji testowej
        private int dimension;
        // liczba populacji - MI
        private int populationSize;
        // liczba osobników w reprodukcji - LAMBDA
        private int reproductionSize;
    }
}
